/*
 * Copyright (c) 2004 Marcus Glocker
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <unistd.h>

/* defines */

#define LINELEN	72			/* maximal encoded line length */
#define EOL	"\r\n"			/* end of line */

/* prototypes */

int base64d(void);
int base64e(void);
int usage(void);

/* data */

static const unsigned char decode[256] = {
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255, 62,255,255,255, 63,
	 52, 53, 54, 55, 56, 57, 58, 59, 60, 61,255,255,255,  0,255,255,
	255,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,255,255,255,255,255,
	255, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};

static const unsigned char encode[64] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*
 * base64d()
 * 	decodes base64 from stdin to plain at stdout
 * Return:
 *	0 = success
 */

int
base64d(void)
{
	int		ch, i;
	unsigned char	in_a[4], in_b[4], out[3];

	while (1) {
		for (i = 0; i < 4; i++) {
			if ((ch = getchar()) == EOF)
				goto quit;

			/* ignore if char is not base64 */
			if (decode[ch] == 255) {
				i--;
				continue;
			}

			/* save it */
			in_a[i] = ch;
			in_b[i] = decode[ch];
		}

		/* split the 24-bit field in 3 8-bit */
		out[0] = (in_b[0] << 2) | (in_b[1] >> 4);
		out[1] = (in_b[1] << 4) | (in_b[2] >> 2);
		out[2] = (in_b[2] << 6) |  in_b[3];

		/* padding */
		i = in_a[2] == '=' ? 1 : (in_a[3] == '=' ? 2 : 3);

		/* print the decoded ascii characters to stdout */
		if (fwrite(out, i, 1, stdout) == EOF)
			exit(1);
	}

	quit:
	return 0;
}

/*
 * base64e()
 * 	encodes plain from stdin to base64 at stdout
 * Return:
 *	0 = success
 */

int
base64e(void)
{
	int		ch, i, j, line = 0;
	unsigned char	in_a[3], out[4];

	while (1) {
		in_a[0] = in_a[1] = in_a[2] = 0;

		for (i = 0; i < 3; i++) {
			if ((ch = getchar()) == EOF)
				break;

			in_a[i] = ch;
		}

		/* split the 24-bit field in 4 6-bit */
		out[0] = encode[(in_a[0] >> 2)];
		out[1] = encode[((in_a[0] & 3) << 4) | (in_a[1] >> 4)];
		out[2] = i < 2 ? '=' : encode[((in_a[1] & 15) << 2) |
			(in_a[2] >> 6)];
		out[3] = i < 3 ? '=' : encode[(in_a[2] & 63)];

		/* limit the line length */
		for (j = 0; j < 4; j++) {
			if (line == LINELEN) {
				fputs(EOL, stdout);
				line = 0;
			}
			putc(out[j], stdout);
			line++;
		}

		/* finished */
		if (i != 3) {
			fputs(EOL, stdout);
			break;
		}
	}

	return 0;
}

/*
 * usage()
 *	prints program usage
 * Return:
 *	0 = success
 */

int
usage(void)
{
	extern char	*__progname;

	printf("usage: %s [-de]\n", __progname);

	return 0;
}

int
main(int argc, char *argv[])
{
	int	ch;

	if (argc < 2) {
		usage();
		exit(1);
	}

	while ((ch = getopt(argc, argv, "de")) != -1) {
		switch (ch) {
			case 'd':
				base64d();
				break;
			case 'e':
				base64e();
				break;
			default:
				usage();
				break;
		}
	}

	return 0;
}
